
#include "libcamera/libcamera.hpp"

#include <alvalue/alvalue.h>
#include <alcommon/alproxy.h>
#include <alcommon/albroker.h>
#include <boost/shared_ptr.hpp>
#include <alcommon/almodule.h>
#include <alcommon/albrokermanager.h>
#include <alcommon/altoolsmain.h>
#include <qi/log.hpp>
#include <sys/mman.h>        /* For shared memory */
#include <fcntl.h>           /* For O_* constants */
#include "utils/speech.hpp"

#include <alvision/alvisiondefinitions.h>


Camera::Camera( boost::shared_ptr<AL::ALBroker> broker, const std::string& name ):
  AL::ALModule(broker, name),
  fVideoProxy(AL::ALVideoDeviceProxy(broker)),
  fGVMId("GVM")
{
  qiLogInfo("camera") << "Constructing libcamera" << std::endl;

  // open shared memory as RW, create if !exist, permissions 600
  shared_fd = shm_open(CAMERA_MEMORY, O_RDWR | O_CREAT, 0600);
  if (shared_fd < 0)
    throw ALERROR(name, "constructor", "shm_open() failed");
  // make shared memory file correct size
  if (ftruncate(shared_fd, sizeof(CameraData)) == -1)
    throw ALERROR(name, "constructor", "ftruncate() failed");
  // map shared memory to process memory
  shared_data = (CameraData*) mmap(NULL, sizeof(CameraData),
                                   PROT_READ | PROT_WRITE,
                                   MAP_SHARED, shared_fd, 0);
  if (shared_data == MAP_FAILED)
    throw ALERROR(name, "constructor", "mmap() failed");
  
  shared_data->readable = 0;
  shared_data->currentCamera = 0;
  shared_data->switchCamera = 0;
  
  pthread_mutexattr_t attr;
  pthread_mutexattr_init(&attr);
  pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_ERRORCHECK);
  pthread_mutex_init(&shared_data->lock, &attr);
  
  // create semaphore with permissions 600, value 1
  semaphore1 = sem_open(CAMERA_SEMAPHORE1, O_RDWR | O_CREAT, 0600, 1);
  if (semaphore1 < 0)
    throw ALERROR(name, "constructor", "sem_open() failed");
  
  // create semaphore with permissions 600, value 0
  semaphore2 = sem_open(CAMERA_SEMAPHORE2, O_RDWR | O_CREAT, 0600, 0);
  if (semaphore2 < 0)
    throw ALERROR(name, "constructor", "sem_open() failed");
    
  missedCycles = 0;
  
  qiLogInfo("camera") << "Constructed libcamera" << std::endl;
}

Camera::~Camera() {
  qiLogInfo("camera") << "Destroying libcamera" << std::endl;
  fCaptureVideo = false;
  pthread_join(fCameraThread, 0);
  fVideoProxy.unsubscribe(fGVMId);
  qi::os::msleep(100);
  if (shared_data != MAP_FAILED) munmap(shared_data, sizeof(CameraData));
  if (shared_fd >= 0) close(shared_fd);
  if (semaphore1 != SEM_FAILED) sem_close(semaphore1);
  if (semaphore2 != SEM_FAILED) sem_close(semaphore2);
  qiLogInfo("camera") << "Destroyed libcamera" << std::endl;
}

void Camera::init() {
  
  // Subscribe
  //fVideoProxy.startFrameGrabber();
  fGVMId = fVideoProxy.subscribe(fGVMId, AL::kVGA, AL::kYUV422ColorSpace, 30);
  qi::os::msleep(100);
  fVideoProxy.setFrameRate(fGVMId, 30);
  fVideoProxy.setParam(AL::kCameraSelectID, 0);
  qi::os::msleep(100);
  
  // Auto Exposition 0-1
  while (fVideoProxy.getParam(AL::kCameraAutoExpositionID) != 0)
    fVideoProxy.setParam(AL::kCameraAutoExpositionID, 0);
  
  // Auto White Balance 0-1
  while (fVideoProxy.getParam(AL::kCameraAutoWhiteBalanceID) != 0)
    fVideoProxy.setParam(AL::kCameraAutoWhiteBalanceID, 0);
  
  // Brightness 0-255
  while (fVideoProxy.getParam(AL::kCameraBrightnessID) != 220) {
    fVideoProxy.setParam(AL::kCameraBrightnessID, 220);
    qiLogInfo("camera") << "brightness: " <<  fVideoProxy.getParam(AL::kCameraBrightnessID) << std::endl;
    qi::os::msleep(10);
  }
  
  // Contrast 0-127
  while (fVideoProxy.getParam(AL::kCameraContrastID) != 50) {
    fVideoProxy.setParam(AL::kCameraContrastID, 50);
    qiLogInfo("camera") << "contrast: " <<  fVideoProxy.getParam(AL::kCameraContrastID) << std::endl;
    qi::os::msleep(10);
  }
  
  // Saturation 0-255
  while (fVideoProxy.getParam(AL::kCameraSaturationID) != 160) {
    fVideoProxy.setParam(AL::kCameraSaturationID, 160);
    qiLogInfo("camera") << "saturation: " <<  fVideoProxy.getParam(AL::kCameraSaturationID) << std::endl;
    qi::os::msleep(10);
  }
  
  // Gain 0-255
  while (fVideoProxy.getParam(AL::kCameraGainID) != 100) {
    fVideoProxy.setParam(AL::kCameraGainID, 100);
    qiLogInfo("camera") << "gain: " <<  fVideoProxy.getParam(AL::kCameraGainID) << std::endl;
    qi::os::msleep(10);
  }
    
  // Horizontal Flip 0-1
  while (fVideoProxy.getParam(AL::kCameraHFlipID) != 1)
    fVideoProxy.setParam(AL::kCameraHFlipID, 1);
  
  // Vertical Flip 0-1
  while (fVideoProxy.getParam(AL::kCameraVFlipID) != 1)
    fVideoProxy.setParam(AL::kCameraVFlipID, 1);
  
    
  // Exposure 0-?
  while (fVideoProxy.getParam(AL::kCameraExposureID) != 20) {
    fVideoProxy.setParam(AL::kCameraExposureID, 20);
    qiLogInfo("camera") << "gain: " <<  fVideoProxy.getParam(AL::kCameraExposureID) << std::endl;
    qi::os::msleep(10);
  }
  
  // White Balance -120 - -36
  while (fVideoProxy.getParam(AL::kCameraWhiteBalanceID) != -80) {
    fVideoProxy.setParam(AL::kCameraWhiteBalanceID, -80);
    qiLogInfo("camera") << "white balance: " <<  fVideoProxy.getParam(AL::kCameraWhiteBalanceID) << std::endl;
    qi::os::msleep(10);
  }
  
  // Auto Exposition 0-1
  while (fVideoProxy.getParam(AL::kCameraAutoExpositionID) != 1)
    fVideoProxy.setParam(AL::kCameraAutoExpositionID, 1);
  
  // Auto White Balance 0-1
  //while (fVideoProxy.getParam(AL::kCameraAutoWhiteBalanceID) != 1)
    //fVideoProxy.setParam(AL::kCameraAutoWhiteBalanceID, 1);
  
  fVideoProxy.setParam(AL::kCameraSelectID, 1);
  qi::os::msleep(100);
  
  // Auto Exposition 0-1
  while (fVideoProxy.getParam(AL::kCameraAutoExpositionID) != 0)
    fVideoProxy.setParam(AL::kCameraAutoExpositionID, 0);
  
  // Auto White Balance 0-1
  while (fVideoProxy.getParam(AL::kCameraAutoWhiteBalanceID) != 0)
    fVideoProxy.setParam(AL::kCameraAutoWhiteBalanceID, 0);
  
  // Brightness 0-255
  while (fVideoProxy.getParam(AL::kCameraBrightnessID) != 220) {
    fVideoProxy.setParam(AL::kCameraBrightnessID, 220);
    qiLogInfo("camera") << "brightness: " <<  fVideoProxy.getParam(AL::kCameraBrightnessID) << std::endl;
    qi::os::msleep(10);
  }
  
  // Contrast 0-127
  while (fVideoProxy.getParam(AL::kCameraContrastID) != 50) {
    fVideoProxy.setParam(AL::kCameraContrastID, 50);
    qiLogInfo("camera") << "contrast: " <<  fVideoProxy.getParam(AL::kCameraContrastID) << std::endl;
    qi::os::msleep(10);
  }
  
  // Saturation 0-255
  while (fVideoProxy.getParam(AL::kCameraSaturationID) != 160) {
    fVideoProxy.setParam(AL::kCameraSaturationID, 160);
    qiLogInfo("camera") << "saturation: " <<  fVideoProxy.getParam(AL::kCameraSaturationID) << std::endl;
    qi::os::msleep(10);
  }
  
  // Gain 0-255
  while (fVideoProxy.getParam(AL::kCameraGainID) != 100) {
    fVideoProxy.setParam(AL::kCameraGainID, 100);
    qiLogInfo("camera") << "gain: " <<  fVideoProxy.getParam(AL::kCameraGainID) << std::endl;
    qi::os::msleep(10);
  }
    
  // Horizontal Flip 0-1
  while (fVideoProxy.getParam(AL::kCameraHFlipID) != 0)
    fVideoProxy.setParam(AL::kCameraHFlipID, 0);
  
  // Vertical Flip 0-1
  while (fVideoProxy.getParam(AL::kCameraVFlipID) != 0)
    fVideoProxy.setParam(AL::kCameraVFlipID, 0);
    
  // Exposure 0-?
  while (fVideoProxy.getParam(AL::kCameraExposureID) != 20) {
    fVideoProxy.setParam(AL::kCameraExposureID, 20);
    qiLogInfo("camera") << "gain: " <<  fVideoProxy.getParam(AL::kCameraExposureID) << std::endl;
    qi::os::msleep(10);
  }
  
  // White Balance -120 - -36
  while (fVideoProxy.getParam(AL::kCameraWhiteBalanceID) != -80) {
    fVideoProxy.setParam(AL::kCameraWhiteBalanceID, -80);
    qiLogInfo("camera") << "white balance: " <<  fVideoProxy.getParam(AL::kCameraWhiteBalanceID) << std::endl;
    qi::os::msleep(10);
  }
  
  // Auto Exposition 0-1
  while (fVideoProxy.getParam(AL::kCameraAutoExpositionID) != 1)
    fVideoProxy.setParam(AL::kCameraAutoExpositionID, 1);
  
  //fVideoProxy.setParam(AL::kCameraSelectID, 0);
  //qi::os::msleep(100);
  
  sem_trywait(semaphore1);
  sem_post(semaphore1);
  fCaptureVideo = true;
  
  pthread_create(&fCameraThread, 0, CameraThread, this);
}



void * Camera::CameraThread( void * pArg )
{
  Camera * lthis = (Camera *) pArg;
  
 /* while ( lthis->fCaptureVideo ) {
    long long start = GetTime();
    lthis->fImagePointer = (AL::ALImage*) lthis->fVideoProxy.getDirectRawImageLocal(lthis->fGVMId);
    //lthis->fImagePointer = (AL::ALImage*) lthis->fVideoProxy.getImageLocal(lthis->fGVMId);
    qiLogInfo("camera") << "raw image took: " <<  (GetTime() - start)/1000 << std::endl;
    start = GetTime();
    //qi::os::msleep(10);
    //qiLogInfo("camera") << "size: " << lthis->fImagePointer->getSize() << std::endl;
    //qiLogInfo("camera") << "size: " << CAMERA_SIZE << std::endl;
    char* dest = (char*)lthis->shared_data->image_data[0];
    char* src = (char*)lthis->fImagePointer->getData();
    unsigned int count = lthis->fImagePointer->getSize();
    /*--dest;
    --src;   
    while (count--) {
      *++dest = *++src;
    }/
    memcpy(dest, src, count);
    qiLogInfo("camera") << "write image took: " <<  (GetTime() - start)/1000 << std::endl;
    //lthis->fVideoProxy.releaseImage(lthis->fGVMId);
    lthis->fVideoProxy.releaseDirectRawImage(lthis->fGVMId);
    //lthis->fVideoProxy.setFrameRate(lthis->fGVMId, 30);
  }*/
  
  
  // Time measurement
  //long long lLastBufferTime = lthis->GetTime();

  // Buffer duration depends on framerate
  //long long lDuration = 1000000000 / 30; // in nanoseconds
  
  while ( lthis->fCaptureVideo )
  {
    /*if((lthis->GetTime() - lLastBufferTime) < lDuration / 1000)
    {
      qi::os::msleep(1);
      continue;
    }*/
    //long long start = lthis->GetTime();
    lthis->GetImage();
    //qiLogInfo("camera") << "GetImage took: " <<  (lthis->GetTime() - start)/1000 << std::endl;
    //lLastBufferTime = lthis->GetTime();
  }
  
  return 0;
}

void Camera::GetImage()
{
  long long start, end;
  int i = 0;
  
  if (shared_data->switchCamera != shared_data->currentCamera) {
    qiLogInfo("camera") << "set camera: " << (int) shared_data->switchCamera << std::endl;
    fVideoProxy.setParam(AL::kCameraSelectID, shared_data->switchCamera);
  }

  // getDirectRawImageLocal
  start = GetTime();
  fImagePointer = (AL::ALImage*) fVideoProxy.getDirectRawImageLocal(fGVMId);
  end = GetTime();
  if (end - start > 40000) qiLogInfo("camera") << "getDirectRawImageLocal took: " <<  (end - start)/1000 << std::endl;
  if (end - start > 100000) SAY("image");
  
  // sem_wait
  //start = GetTime();
  sem_wait(semaphore1);
  //end = GetTime();
  //if (end - start > 40000) qiLogInfo("camera") << "sem_wait took: " <<  (end - start)/1000 << std::endl;
  
  // memcpy
  //start = GetTime();
  memcpy(shared_data->image_data[!shared_data->readable], fImagePointer->getData(), fImagePointer->getSize());
  //end = GetTime();
  //if (end - start > 40000) qiLogInfo("camera") << "memcpy took: " <<  (end - start)/1000 << std::endl;
  
  shared_data->currentCamera = fImagePointer->getCameraId();
  
  sem_post(semaphore1);
  fVideoProxy.releaseDirectRawImage(fGVMId);
  
  // Check if warriors got image
  sem_getvalue(semaphore2, &i);
  if (i > 0) {
    if (missedCycles < 2) {
      qiLogError("camera") << "libcamera: warriors missed camera cycle" << std::endl;
    }
    missedCycles++;
  } else {
    if (missedCycles >= 2) {
      qiLogError("camera") << "libcamera: warriors returned to camera cycle" << std::endl;
    }
    missedCycles = 0;
    sem_post(semaphore2);
  }
 
}

/*void Camera::GetImage() {
  long long write = GetTime();
  fImagePointer = (AL::ALImage*) fVideoProxy.getDirectRawImageLocal(fGVMId);
  qiLogInfo("camera") << "raw image took: " <<  (GetTime() - write)/1000 << std::endl;

  //long long timeStamp = fImagePointer->getTimeStamp();
  const unsigned char* dataPointer = fImagePointer->getData();
  unsigned int size = fImagePointer->getSize();
  
  //qiLogInfo("camera") << timeStamp << std::endl;
  
  int i = 0;
  //sem_getvalue(semaphore2, &i);
  //qiLogInfo("camera") << i << std::endl;
  if (i <= 0) {
    if (missedCycles >= 2) {
      qiLogInfo("camera") << "warriors returned to camera cycle" << std::endl;
    }
    missedCycles = 0;
    write = GetTime();
    sem_wait(semaphore1);
    qiLogInfo("camera") << "sem_wait took: " <<  (GetTime() - write)/1000 << std::endl;
    write = GetTime();
    memcpy(shared_data->image_data[!shared_data->readable], dataPointer, size);
    qiLogInfo("camera") << "write image took: " <<  (GetTime() - write)/1000 << std::endl;
    sem_post(semaphore1);
    //sem_post(semaphore2);
  } else {
    if (missedCycles < 2) {
      qiLogError("camera") << "warriors missed camera cycle" << std::endl;
    }
    missedCycles++;
  }

  /** Do not forget to release the image. /
  fVideoProxy.releaseDirectRawImage(fGVMId);
}*/

long long Camera::GetTime() {
  qi::os::timeval lTimeStruct;
  qi::os::gettimeofday(&lTimeStruct);
  long long lTime = (long long) lTimeStruct.tv_sec * 1000000; // Seconds
  lTime += (long long) lTimeStruct.tv_usec;                   // Microseconds
  return lTime;
}

extern "C"
{
  int _createModule(boost::shared_ptr<AL::ALBroker> pBroker)
  {
    AL::ALBrokerManager::setInstance(pBroker->fBrokerManager.lock());
    AL::ALBrokerManager::getInstance()->addBroker(pBroker);
    AL::ALModule::createModule<Camera>( pBroker, "Camera" );
    return 0;
  }

  int _closeModule()
  {
    return 0;
  }
}

