
#include <cmath>
#include <iostream>
#include <string>
#include <vector>
#include "KalmanFilter.hpp"
#include "utils/angles.hpp"

#define THE_GOAL_WIDTH 1400.0
#define THE_FIELD_LENGTH 6000.0
#define THE_FIELD_WIDTH 4000.0

//Noise Covariance Matrix
#define NOISE_COV_X .002
#define NOISE_COV_Y .002
#define NOISE_COV_THETA .002

//Measurement Covariance Matrices
#define POST_ONE_COV_X .5
#define POST_ONE_COV_Y .5
#define POST_ONE_COV_THETA .5
#define POST_BOTH_COV_X .1
#define POST_BOTH_COV_Y .1
#define POST_BOTH_COV_THETA .1

//Mathematical Constants
#define PI 3.1415927

//Organizational Constants
#define X 0
#define Y 1
#define THETA 2

using namespace std;

/**
* Constructor; initializes all matrices needed for the Kalman filter
*/
KalmanFilter::KalmanFilter() {
    initialized = false;

	for (int i = 0; i < 3; i++)
    {
        for (int j = 0; j < 3; j++)
        {
            P[i][j] = 0;
			Q[i][j] = 0;
			R1[i][j] = 0;
			R2[i][j] = 0;
			J[i][j] = 0;
        }
    }

	Q[0][0] = NOISE_COV_X;
	Q[1][1] = NOISE_COV_Y;
	Q[2][2] = NOISE_COV_THETA;
	
	R1[0][0] = POST_ONE_COV_X;
	R1[1][1] = POST_ONE_COV_Y;
	R1[2][2] = POST_ONE_COV_THETA;
	
	R2[0][0] = POST_BOTH_COV_X;
	R2[1][1] = POST_BOTH_COV_Y;
	R2[2][2] = POST_BOTH_COV_THETA;
	
	P[0][0] = .5;
	P[1][1] = .5;
	P[2][2] = .5;
}

/** 
* Localization update function to be called on every tick of localization
* return the robot's position after the update
*/
Position KalmanFilter::localise() {
    odometryUpdate();
    postLocalise();
    prev_odometry = odometry;
    
    robotPos[0].var[0] = P[0][0];
    robotPos[0].var[1] = P[1][1];
    robotPos[0].var[2] = P[2][2];
    
    if (initialized == false)
    {
        initialized = true;
    }

    //bounds checking
    if (robotPos[0].x > THE_FIELD_WIDTH / 2.0 + 500)
    {
        robotPos[0].x = THE_FIELD_WIDTH / 2.0 + 500;
        P[0][0] += .25;
    }
    else if (robotPos[0].x < -(THE_FIELD_WIDTH / 2.0 + 500))
    {
        robotPos[0].x = -(THE_FIELD_WIDTH / 2.0 + 500);
        P[0][0] += .25;
    }
    if (robotPos[0].y > THE_FIELD_LENGTH / 2.0 + 500)
    {
        robotPos[0].y = THE_FIELD_LENGTH / 2.0 + 500;
        P[1][1] += .25;
    }
    else if (robotPos[0].y < -(THE_FIELD_LENGTH / 2.0 + 500))
    {
        robotPos[0].y = -(THE_FIELD_LENGTH / 2.0 + 500);
        P[1][1] += .25;
    }

    //NOTE: experimental code to adjust how quickly new positions are jumped to
    if (P[0][0] < .25)
        P[0][0] = .25;
    if (P[1][1] < .25)
        P[1][1] = .25;
    if (P[2][2] < .25)
        P[2][2] = .25;

    return *(new Position(robotPos[0].x, robotPos[0].y, robotPos[0].theta, 
        robotPos[0].var[0] + robotPos[0].var[1] + robotPos[0].var[2]));
}

/**
* Post update which is a measurement update if the robot sees either one or both goal posts
* of either color
*/
void KalmanFilter::postLocalise() {
	
	double y[3]; //measurement vector
	
    if (obsWhichPosts == pBLUE_BOTH || obsWhichPosts == pYELLOW_BOTH)
    {
        //initialization
        double d1 = obsPosts[0].distance;   //distance to left post
        double d2 = obsPosts[1].distance;   //distance to right post
        double d3 = THE_GOAL_WIDTH;
        //double t1 = obsPosts[0].heading;  //angle of left post with respect to robot (currently unused)
        double t2 = obsPosts[1].heading;    //angle of right post with respect to robot
        
        //calculate position and orientation
        /*
        Note: temp is the input to the acos function as determined by the law of cosines.
        The triangle side lengths are determined by the distance between the goal posts 
        (which is both constant and known), and the robot's perceived distance
        to each post (which can often be faulty).  Due to these faulty measurements, it's
        possible to get triangle side lengths that cannot physically make a triangle, so temp
        falls outside of the -1 to 1 range of inputs to acos.  The current solution implemented
        here simply divides the value by 2 until it falls into that range.  This both keeps the
        sign correct (so the robot is still on the correct side of the posts, as this is
        almost always perceived correctly) and usually keeps the error under 45 degrees.
        */
        double alpha;
        double temp = ((d1*d1)-(d2*d2)-(d3*d3))/(-2.0*d2*d3);
        while (temp < -1.0 || temp > 1.0)
        {
            temp /= 2.0;
        }

        alpha = acos(temp);
		y[THETA] = alpha - t2;
		y[THETA] = correctAngle(y[THETA]);
        y[X] = THE_GOAL_WIDTH/2.0 - d2*cos(alpha);
        y[Y] = THE_FIELD_LENGTH/2.0 - d2*sin(alpha);
        
        
        //correction for when the goal is the friendly goal
        if ((obsWhichPosts == pYELLOW_BOTH && obsTeamRed == true) || (obsWhichPosts == pBLUE_BOTH && obsTeamRed == false))
        {
            y[THETA] = y[THETA] + 3.14159;
            y[THETA] = correctAngle(y[THETA]);
            y[X] = -y[X];
            y[Y] = -y[Y];
        }
		
		//Kalman Filter Measurement Update Equations
		//Calculate Kalman Gain
		J[0][0] = P[0][0]*(1.0 / (P[0][0] + R2[0][0]));
		J[1][1] = P[1][1]*(1.0 / (P[1][1] + R2[1][1]));
		J[2][2] = P[2][2]*(1.0 / (P[2][2] + R2[2][2]));
		
		//Calculate new position using Kalman Gain
		robotPos[0].x = robotPos[0].x + J[0][0] * (y[X] - robotPos[0].x);
		robotPos[0].y = robotPos[0].y + J[1][1] * (y[Y] - robotPos[0].y);
		robotPos[0].theta = robotPos[0].theta + J[2][2] * (y[THETA] - robotPos[0].theta);
		
		robotPos[0].theta = correctAngle(robotPos[0].theta);
		
		//Calculate updated error covariance matrix
		P[0][0] = (1.0 - J[0][0]) * P[0][0];
		P[1][1] = (1.0 - J[1][1]) * P[1][1];
		P[2][2] = (1.0 - J[2][2]) * P[2][2];
    }
    else if (obsWhichPosts == pBLUE_LEFT || obsWhichPosts == pBLUE_RIGHT || obsWhichPosts == pYELLOW_LEFT || obsWhichPosts == pYELLOW_RIGHT)
    {
        
        //initialization
        double possibleLocations[181][2];
        double angle = 0.0;
        double d = obsPosts[0].distance;
        double t = obsPosts[0].heading;
        int xSign = 1, ySign = 1;
        
        //populate possible locations (arc around the post)
        if ((obsWhichPosts == pBLUE_LEFT && obsTeamRed == true) || (obsWhichPosts == pYELLOW_LEFT && obsTeamRed == false))
        {
            xSign = -1;
            ySign = 1;
        }
        else if ((obsWhichPosts == pBLUE_RIGHT && obsTeamRed == true) || (obsWhichPosts == pYELLOW_RIGHT && obsTeamRed == false))
        {
            xSign = 1;
            ySign = 1;
        }
        else if ((obsWhichPosts == pYELLOW_LEFT && obsTeamRed == true) || (obsWhichPosts == pBLUE_LEFT && obsTeamRed == false))
        {
            xSign = 1;
            ySign = -1;
        }
        else if ((obsWhichPosts == pYELLOW_RIGHT && obsTeamRed == true) || (obsWhichPosts == pBLUE_RIGHT && obsTeamRed == false))
        {
            xSign = -1;
            ySign = -1;
        }
        
        for (int i=0; i<=180; i++)
        {
            possibleLocations[i][0] = xSign*(700 - d*cos(angle));
            possibleLocations[i][1] = ySign*(3000 - d*sin(angle));
            angle += 0.0174532925;
        }
        
        //find the closest location to the previous (x,y)
        double minDist = 999999.0;
        double minCoord[2];
        double dist;
        minCoord[0] = 0.0;
        minCoord[1] = 0.0;
        for (int j=0; j<180; j++)
        {
            dist = sqrt((robotPos[0].x-possibleLocations[j][0])*(robotPos[0].x-possibleLocations[j][0]) + (robotPos[0].y-possibleLocations[j][1])*(robotPos[0].y-possibleLocations[j][1]));
            if (dist < minDist)
            {
                minDist = dist;
                minCoord[0] = possibleLocations[j][0];
                minCoord[1] = possibleLocations[j][1];
            }
        }
        y[X] = minCoord[0];
        y[Y] = minCoord[1];
        y[THETA] = atan2(y[Y], y[X]) - t;
        
        y[THETA] = correctAngle(y[THETA]);
        
        //Kalman Filter Measurement Update Equations
		//Calculate Kalman Gain
		J[0][0] = P[0][0]*(1.0 / (P[0][0] + R1[0][0]));
		J[1][1] = P[1][1]*(1.0 / (P[1][1] + R1[1][1]));
		J[2][2] = P[2][2]*(1.0 / (P[2][2] + R1[2][2]));
		
		//Calculate new position using Kalman Gain
		robotPos[0].x = robotPos[0].x + J[0][0] * (y[X] - robotPos[0].x);
		robotPos[0].y = robotPos[0].y + J[1][1] * (y[Y] - robotPos[0].y);
		robotPos[0].theta = robotPos[0].theta + J[2][2] * (y[THETA] - robotPos[0].theta);
		
		robotPos[0].theta = correctAngle(robotPos[0].theta);
		
		//Calculate updated error covariance matrix
		P[0][0] = (1.0 - J[0][0]) * P[0][0];
		P[1][1] = (1.0 - J[1][1]) * P[1][1];
		P[2][2] = (1.0 - J[2][2]) * P[2][2];
    }
}

/**
* Odometry update which is a time update that accounts for any movement since the previous 
* update
*/
void KalmanFilter::odometryUpdate() {
    if (initialized == true)
    {
        double u[3];    //change in odometry
        double B[3][3]; //transformation matrix

        //Get odometry change
        Odometry change = odometry - prev_odometry;
        u[X] = correctLeft(change.left);
        u[Y] = correctForward(change.forward);
        u[THETA] = correctTurn(change.turn);

        //Setup transformation matrix B to transform the odometry update from local
        //coordinates to global coordinates
        B[0][0] = -cos(PI/2.0 - robotPos[0].theta);
        B[0][1] = sin(PI/2.0 - robotPos[0].theta);
        B[0][2] = 0.0;
        B[1][0] = -cos(PI - robotPos[0].theta);
        B[1][1] = sin(PI - robotPos[0].theta);
        B[1][2] = 0.0;
        B[2][0] = 0.0;
        B[2][1] = 0.0;
        B[2][2] = 1.0;
        
        //Kalman Filter Time Update Equations
        //Update position based on odometry
        robotPos[0].x += u[X]*B[0][0] + u[Y]*B[1][0] + u[THETA]*B[2][0];
        robotPos[0].y += u[X]*B[0][1] + u[Y]*B[1][1] + u[THETA]*B[2][1];
        robotPos[0].theta += u[X]*B[0][2] + u[Y]*B[1][2] + u[THETA]*B[2][2];
        
        robotPos[0].theta = correctAngle(robotPos[0].theta);
        
        //Update error covariance matrix
        /*
        Note: errors are only updated when the odometry is non-zero.  A non-zero odometry
        for the x direction or the y direction requires updates to both the x and y error,
        since the odometry is calculated with respect to the local coordinate frame of the
        robot.
        */
        if (u[X] != 0 || u[Y] != 0)
        {
            P[0][0] += Q[0][0];
            P[1][1] += Q[1][1];
        }
        if (u[THETA] != 0)
        {
            P[2][2] += Q[2][2];
        }
    }
}

/**
* Fixes angles that move outside of the bounds of [0, 2*PI)
* param angle the angle to be checked
* return an equivalent angle within the bounds
*/
double KalmanFilter::correctAngle(double angle) {
    while (angle >= 2*PI)
    {
        angle -= 2*PI;
    }
    while (angle < 0.0)
    {
        angle += 2*PI;
    }
    
    return angle;
}

