
#include <cmath>
#include <iostream>
#include <string>
#include <vector>
#include "RobotLocaliser.hpp"
#include "utils/angles.hpp"

#define THE_GOAL_WIDTH 1400.0
#define THE_FIELD_LENGTH 6000.0

using namespace std;

RobotLocaliser::RobotLocaliser() {
    initialized = false;
}

bool RobotLocaliser::localise() {
    odometryUpdate();
    postLocalise();
    prev_odometry = odometry;
    if (initialized == false)
    {
        initialized = true;
    }
    return true;
}

void RobotLocaliser::postLocalise() {
    // Calculate position and orientation from two posts
    if (obsWhichPosts == pBLUE_BOTH || obsWhichPosts == pYELLOW_BOTH)
    {
        //initialization
        double d1 = obsPosts[0].distance;
        double d2 = obsPosts[1].distance;
        double d3 = THE_GOAL_WIDTH;
        //double t1 = obsPosts[0].heading;  //currently unused
        double t2 = obsPosts[1].heading;
        
        //calculate position and orientation
        double alpha = acos(((d1*d1)-(d2*d2)-(d3*d3))/(-2.0*d2*d3));
        robotPos[0].theta = alpha + t2;
        robotPos[0].x = THE_GOAL_WIDTH/2.0 - d2*cos(alpha);
        robotPos[0].y = THE_FIELD_LENGTH/2.0 - d2*sin(alpha);
        
        //correction for opposite side of field
        if (obsWhichPosts == pYELLOW_BOTH);
        {
            robotPos[0].theta = robotPos[0].theta + 3.14159;
            robotPos[0].x = -robotPos[0].x;
            robotPos[0].y = -robotPos[0].y;
        }
        
        robotPos[0].var[0] = 0.0;
    }
    
    // Update position from one post
    else if (obsWhichPosts == pBLUE_LEFT || obsWhichPosts == pBLUE_RIGHT || obsWhichPosts == pYELLOW_LEFT || obsWhichPosts == pYELLOW_RIGHT)
    {
        
        //initialization
        double possibleLocations[181][2];
        double angle = 0;
        double d = obsPosts[0].distance;
        double t = obsPosts[0].heading;
        int xSign = 1, ySign = 1;
        
        //populate possible locations (arc around the post)
        if (obsWhichPosts == pBLUE_LEFT)
        {
            xSign = -1;
            ySign = 1;
        }
        else if (obsWhichPosts == pBLUE_RIGHT)
        {
            xSign = 1;
            ySign = 1;
        }
        else if (obsWhichPosts == pYELLOW_LEFT)
        {
            xSign = 1;
            ySign = -1;
        }
        else if (obsWhichPosts == pYELLOW_RIGHT)
        {
            xSign = -1;
            ySign = -1;
        }
        for (int i=0; i<=180; i++)
        {
            possibleLocations[i][0] = xSign*(700 - d*cos(angle));
            possibleLocations[i][1] = ySign*(3000 - d*sin(angle));
            angle += 0.0174532925;
        }
        
        //find the closest location to the previous (x,y)
        double minDist = 999999.0;
        double minCoord[2];
        double dist;
        minCoord[0] = 0.0;
        minCoord[1] = 0.0;
        for (int j=0; j<180; j++)
        {
            dist = sqrt((robotPos[0].x-possibleLocations[j][0])*(robotPos[0].x-possibleLocations[j][0]) + (robotPos[0].y-possibleLocations[j][1])*(robotPos[0].y-possibleLocations[j][1]));
            if (dist < minDist)
            {
                minDist = dist;
                minCoord[0] = possibleLocations[j][0];
                minCoord[1] = possibleLocations[j][1];
            }
        }
        robotPos[0].x = minCoord[0];
        robotPos[0].y = minCoord[1];
        robotPos[0].theta = atan2(robotPos[0].y, robotPos[0].x) + t;
        
        robotPos[0].var[0] += 0.001;
        if (robotPos[0].var[0] > 1.0)
            robotPos[0].var[0] = 1.0;
    }
    
}

void RobotLocaliser::odometryUpdate() {
    if (initialized == true)
    {
        Odometry change = odometry - prev_odometry;
        float changeForward = change.forward;
        float changeLeft = change.left;
        float changeTurn = change.turn;
        
        robotPos[0].x = robotPos[0].x 
            - changeLeft*cos(3.14159/2.0 - robotPos[0].theta)
            - changeForward*cos(3.14159 - robotPos[0].theta);
        
        robotPos[0].y = robotPos[0].y 
            + changeLeft*sin(3.14159/2.0 - robotPos[0].theta)
            + changeForward*sin(3.14159 - robotPos[0].theta);
        
        robotPos[0].theta = robotPos[0].theta + changeTurn;
        while (robotPos[0].theta > 3.14159*2.0)
        {
            robotPos[0].theta = robotPos[0].theta - 3.14159*2.0;
        }
        while (robotPos[0].theta < 0.0)
        {
            robotPos[0].theta = robotPos[0].theta + 3.14159*2.0;
        }
        
        if (changeForward > 0.0)
            robotPos[0].var[0] += 0.0025;
        if (changeLeft > 0.0)
            robotPos[0].var[0] += 0.0025;
        if (changeTurn > 0.0)
            robotPos[0].var[0] += 0.0025;
        if (robotPos[0].var[0] > 1.0)
            robotPos[0].var[0] = 1.0;
    }
}

