/*
 * ParticleFilter.cpp
 *
 *  Created on: Feb 5, 2012
 *      Author: Ryan
 */

#include "ParticleFilter.hpp"
#include "perception/localisation/OdometryCorrection.hpp"
#include <stdlib.h>
#include <math.h>

using namespace std;

ParticleFilter::ParticleFilter(){
	initialize(-1, NULL);
}

ParticleFilter::ParticleFilter(int num_particles){
	initialize(num_particles, NULL);
}

ParticleFilter::~ParticleFilter(){
	free(particles.positions);
}

void ParticleFilter::initialize(int num_particles, pf_def* i_def){
	penalizeProc = false;

	if(i_def == NULL){
		def.operation.high_prob_percentage = DEFAULT_HIGH_PROB_PERCENTAGE;
		def.operation.turnover_percentage = DEFAULT_TURNOVER_PERCENTAGE;

		def.tolerance.point_bound = DEFAULT_POINT_TOL;
		def.tolerance.theta_bound = DEFAULT_THETA_TOL;
		def.tolerance.max_region_prob_dif = DEFAULT_MAX_REGION_PROB_DIF;

		def.prob_factor.init = DEFAULT_INIT_PROB_ADJUST;
		def.prob_factor.direction = DEFAULT_DIRECTION_PROB_MOD;
		def.prob_factor.either_post = DEFAULT_EITHER_POST_PROB_ADJUST;
		def.prob_factor.invalid_triangle = DEFAULT_INVALID_TRI_PROB_ADJUST;
		def.prob_factor.one_post = DEFAULT_ONE_POST_PROB_ADJUST;
		def.prob_factor.unknown_posts = DEFAULT_UNKNOWN_POSTS_PROB_ADJUST;
		def.prob_factor.odometry = DEFAULT_ODOMETRY_PROB_MOD;
	}else{
		def.operation.high_prob_percentage = i_def->operation.high_prob_percentage;
		def.operation.turnover_percentage = i_def->operation.turnover_percentage;

		def.tolerance.point_bound = i_def->tolerance.point_bound;
		def.tolerance.theta_bound = i_def->tolerance.theta_bound;
		def.tolerance.max_region_prob_dif = i_def->tolerance.max_region_prob_dif;

		def.prob_factor.init = i_def->prob_factor.init;
		def.prob_factor.direction = i_def->prob_factor.direction;
		def.prob_factor.either_post = i_def->prob_factor.either_post;
		def.prob_factor.invalid_triangle = i_def->prob_factor.invalid_triangle;
		def.prob_factor.one_post = i_def->prob_factor.one_post;
		def.prob_factor.unknown_posts = i_def->prob_factor.unknown_posts;
		def.prob_factor.odometry = i_def->prob_factor.odometry;
	}

	orig_def.operation.high_prob_percentage = def.operation.high_prob_percentage;
	orig_def.operation.turnover_percentage = def.operation.turnover_percentage;

	orig_def.tolerance.point_bound = def.tolerance.point_bound;
	orig_def.tolerance.theta_bound = def.tolerance.theta_bound;
	orig_def.tolerance.max_region_prob_dif = def.tolerance.max_region_prob_dif;

	orig_def.prob_factor.init = def.prob_factor.init;
	orig_def.prob_factor.direction = def.prob_factor.direction;
	orig_def.prob_factor.either_post = def.prob_factor.either_post;
	orig_def.prob_factor.invalid_triangle = def.prob_factor.invalid_triangle;
	orig_def.prob_factor.one_post = def.prob_factor.one_post;
	orig_def.prob_factor.unknown_posts = def.prob_factor.unknown_posts;
	orig_def.prob_factor.odometry = def.prob_factor.odometry;

	if(num_particles <= 0){
		particles.num_pos = NUM_PARTICLES;
	}else{
		particles.num_pos = num_particles;
	}

	particles.positions = (pos*)malloc(particles.num_pos*sizeof(pos));

	for(int i = 0; i < particles.num_pos; i++){
		particles.positions[i].x = 0;
		particles.positions[i].y = 0;
		particles.positions[i].theta = 0;
		particles.positions[i].prob = TRIMMED_PARTICLE_PROB;
	}

	//Populate particles (createParticles)
	bounds field;
	field.average_prob = TRIMMED_PARTICLE_PROB;
	field.x_min = -1*FIELD_MM_X/2;
	field.x_max = FIELD_MM_X/2;
	field.y_min = -1*FIELD_MM_Y/2;
	field.y_max = FIELD_MM_Y/2;
	field.theta_min = MIN_THETA;
	field.theta_max = MAX_THETA;

	bounds_array tofill;
	tofill.num_bounds = 1;
	tofill.bounds_list = (bounds*)malloc(sizeof(bounds));
	tofill.bounds_list[0] = field;

	createParticles(particles.num_pos, &particles, tofill);

	//Set up odometry tracking
	prevOdo.forward = 0;
	prevOdo.left = 0;
	prevOdo.turn = 0;

	free(tofill.bounds_list);
}


Position ParticleFilter::localise(){
	Position retPos(0,0,0,0);

	int penalty = our_team.players[player_number].penalty;

	if((!penalizeProc)&&(penalty != PENALTY_NONE)){
		cout << "PENALIZED! Redistributing particles" << endl;

		//Populate particles (createParticles)
		bounds field;
		field.average_prob = TRIMMED_PARTICLE_PROB;
		field.x_min = -1*FIELD_MM_X/2;
		field.x_max = FIELD_MM_X/2;
		field.y_min = -1*FIELD_MM_Y/2;
		field.y_max = 0; 				//Always enter on our side of the field
		field.theta_min = MIN_THETA;
		field.theta_max = MAX_THETA;

		bounds_array tofill;
		tofill.num_bounds = 1;
		tofill.bounds_list = (bounds*)malloc(sizeof(bounds));
		tofill.bounds_list[0] = field;

		createParticles(particles.num_pos, &particles, tofill);

		free(tofill.bounds_list);

		penalizeProc = true;
	}else{
		retPos = update();

		robotPos[0].x = retPos.x;
		robotPos[0].y = retPos.y;
		robotPos[0].theta = retPos.theta;

		double* vars = calculateVariance(getHighProbParticles(particles, def.operation.high_prob_percentage));

		robotPos[0].var[0] = retPos.probability;
		robotPos[0].var[1] = vars[0]*vars[1];
		robotPos[0].var[2] = vars[2];

		penalizeProc = false;
	}



	cout << retPos.x << "," << retPos.y << "," << retPos.theta << "," << retPos.probability << endl;

	return retPos;
}

Position ParticleFilter::update(){
	adjustParticles(&particles);

	if(evaluateParticles(&particles)){
		normalizeParticleProbability(&particles);

		sortParticles(&particles);

		trimParticles(&particles, def.operation.turnover_percentage);
	}

	normalizeParticleProbability(&particles);

	//normalizeProbabilityCenter(&particles);

	multiplyParticles(&particles);

	sortParticles(&particles);

	//Normalize all particle theta
	for(int i = 0; i < particles.num_pos; i++){
		particles.positions[i].theta = normalizeTheta(particles.positions[i].theta);
	}

	Position retPos(particles.positions[0].x,
			particles.positions[0].y,
			particles.positions[0].theta,
			particles.positions[0].prob);

	return retPos;
}

void ParticleFilter::adjustParticles(pos_array* part){
	odometryAdjustment(odometry, part);

	//Kill probability of all particles outside bounds
	for(int i = 0; i < particles.num_pos; i++){
		if(!(inBounds(particles.positions[i]))){
			particles.positions[i].prob = TRIMMED_PARTICLE_PROB;
		}
	}
}

bool ParticleFilter::evaluateParticles(pos_array* part){
	bool changed = false;

	//Process sensor data
	goalheur_precalc_data ghp_data = goalPostHeuristicPreCalc(obsWhichPosts, obsPosts, obsTeamRed);
	//lineangle_precalc_data lahp_data = lineAngleHeuristicPreCalc(numSeenEdges, seenEdges);

	if(ghp_data.iff != P_NO_DATA){
		for(int i = 0; i < part->num_pos; i++){
			goalPostHeuristic(&part->positions[i], ghp_data);
			//lineAngleHeuristic(&part->positions[i], lahp_data);
		}

		changed = true;
	}

	free(ghp_data.data);

	return changed;
}

void ParticleFilter::trimParticles(pos_array* part, double percentage){
	int num_to_trim = percentage*part->num_pos;

	if(num_to_trim <= 0){
		return;
	}

	sortParticles(part);

	for(int i = (part->num_pos - num_to_trim); i < part->num_pos; i++){
		part->positions[i].prob = TRIMMED_PARTICLE_PROB;
	}
}

void ParticleFilter::multiplyParticles(pos_array* part){
	int num_trimmed = 0;

	for(int i = 0; i < part->num_pos; i++){
		if(part->positions[i].prob == TRIMMED_PARTICLE_PROB){
			num_trimmed++;
		}
	}

	pos_array high_prob = getHighProbParticles(*part, def.operation.high_prob_percentage);

	bounds_array gen_bounds = determineBounds(high_prob);

	createParticles(num_trimmed, part, gen_bounds);

	free(high_prob.positions);
	free(gen_bounds.bounds_list);
}

void ParticleFilter::sortParticles(pos_array* part){
	//If already sorted, return
	if(sorted(*part)){return;}

	pos tempPos;

	//Go through the array from index 0, switching the element and shifting if necessary
	for(int curUnsortedHighest = 0; curUnsortedHighest < part->num_pos-1; curUnsortedHighest++){
		for(int comparing = curUnsortedHighest+1; comparing < part->num_pos; comparing++){
			if(part->positions[comparing].prob > part->positions[curUnsortedHighest].prob){
				tempPos = part->positions[comparing];

				//Shift particles one space down
				for(int shifting = comparing; shifting > curUnsortedHighest; shifting--){
					part->positions[shifting] = part->positions[(shifting-1)];
				}

				//replace original
				part->positions[curUnsortedHighest] = tempPos;
			}
		}
	}
}

bool ParticleFilter::sorted(pos_array part){
	for(int i = 1; i < part.num_pos; i++){
		if(part.positions[i-1].prob < part.positions[i].prob){
			return false;
		}
	}

	return true;
}

bool ParticleFilter::inBounds(pos position){
	if((position.x > (FIELD_MM_X/2))||(position.x < (-1*FIELD_MM_X/2))){return false;}

	if((position.y > (FIELD_MM_Y/2))||(position.y < (-1*FIELD_MM_Y/2))){return false;}

	return true;

}

void ParticleFilter::createParticles(int num_particles, pos_array* part_array, bounds_array bounds_to_fill){
	if(bounds_to_fill.num_bounds < 1){
		return;
	}

	if(bounds_to_fill.num_bounds > 1){
		if(bounds_to_fill.num_bounds > num_particles){
			//Find "least" average probability among bounds and eliminate, recalling this function
			int leastIndex = 0;

			for(int i = 1; i < bounds_to_fill.num_bounds; i++){
				if(bounds_to_fill.bounds_list[i].average_prob < bounds_to_fill.bounds_list[leastIndex].average_prob){
					leastIndex = i;
				}
			}

			bounds_array new_bounds_to_fill;
			new_bounds_to_fill.bounds_list = new bounds[bounds_to_fill.num_bounds - 1];
			new_bounds_to_fill.num_bounds = bounds_to_fill.num_bounds - 1;

			int newIndex = 0;

			for(int i = 0; i < bounds_to_fill.num_bounds; i++){
				if(i != leastIndex){
					new_bounds_to_fill.bounds_list[newIndex] = bounds_to_fill.bounds_list[i];
					newIndex++;
				}
			}

			createParticles(num_particles, part_array, new_bounds_to_fill);

			return;
		}

		//Check minimums and seperate
		int part[bounds_to_fill.num_bounds];
		int part_assigned = 0;

		for(int i = 0; i < bounds_to_fill.num_bounds; i++){
			double xDiff, yDiff, tDiff;

			//Check that max and min are correct - if not, switch using XOR trick
			if(bounds_to_fill.bounds_list[i].x_min > bounds_to_fill.bounds_list[i].x_max){
				double temp = bounds_to_fill.bounds_list[i].x_min;
				bounds_to_fill.bounds_list[i].x_min = bounds_to_fill.bounds_list[i].x_max;
				bounds_to_fill.bounds_list[i].x_max = temp;
			}

			if(bounds_to_fill.bounds_list[i].y_min > bounds_to_fill.bounds_list[i].y_max){
				double temp = bounds_to_fill.bounds_list[i].y_min;
				bounds_to_fill.bounds_list[i].y_min = bounds_to_fill.bounds_list[i].y_max;
				bounds_to_fill.bounds_list[i].y_max = temp;
			}

			if(bounds_to_fill.bounds_list[i].theta_min > bounds_to_fill.bounds_list[i].theta_max){
				double temp = bounds_to_fill.bounds_list[i].theta_min;
				bounds_to_fill.bounds_list[i].theta_min = bounds_to_fill.bounds_list[i].theta_max;
				bounds_to_fill.bounds_list[i].theta_max = temp;
			}

			xDiff = bounds_to_fill.bounds_list[i].x_max - bounds_to_fill.bounds_list[i].x_min;
			yDiff = bounds_to_fill.bounds_list[i].y_max - bounds_to_fill.bounds_list[i].y_min;
			tDiff = bounds_to_fill.bounds_list[i].theta_max - bounds_to_fill.bounds_list[i].theta_min;

			part[i] = (int)((xDiff/FIELD_MM_X)*(yDiff/FIELD_MM_Y)*(tDiff/(MAX_THETA - MIN_THETA))*num_particles);
			part_assigned += part[i];
		}

		if(part_assigned < num_particles){
			for(int i = 0; i < bounds_to_fill.num_bounds; i++){
				if(part[i] == 0){
					part[i]++;
					part_assigned++;
				}

				if(part_assigned == num_particles){
					break;
				}
			}

			if(part_assigned < num_particles){
				part[0] += (num_particles - part_assigned);
			}
		}

		bounds_array new_bounds;
		new_bounds.num_bounds = 1;
		new_bounds.bounds_list = new bounds[1];

		for(int i = 0; i < bounds_to_fill.num_bounds; i++){
			new_bounds.bounds_list[0] = bounds_to_fill.bounds_list[i];
			createParticles(part[i], part_array, new_bounds);
		}

		return;
	}

	//Process single bounds
	bounds current_bounds = bounds_to_fill.bounds_list[0];

	double xDiff, yDiff, tDiff;

	//Check that max and min are correct - if not, switch using XOR trick
	if(current_bounds.x_min > current_bounds.x_max){
		double temp = current_bounds.x_min;
		current_bounds.x_min = current_bounds.x_max;
		current_bounds.x_max = temp;
	}

	if(current_bounds.y_min > current_bounds.y_max){
		double temp = current_bounds.y_min;
		current_bounds.y_min = current_bounds.y_max;
		current_bounds.y_max = temp;
	}

	if(current_bounds.theta_min > current_bounds.theta_max){
		double temp = current_bounds.theta_min;
		current_bounds.theta_min = current_bounds.theta_max;
		current_bounds.theta_max = temp;
	}

	xDiff = current_bounds.x_max - current_bounds.x_min;
	yDiff = current_bounds.y_max - current_bounds.y_min;
	tDiff = current_bounds.theta_max - current_bounds.theta_min;

	divs_struct divs = determineDivisions(num_particles, xDiff, yDiff, tDiff);

	double startingTheta, thetaStep;

	if(divs.tDiv == tDiff){
		startingTheta = tDiff/2 + current_bounds.theta_min;
		thetaStep = divs.tDiv;
	}else if(divs.tDiv == tDiff/2){
		startingTheta = current_bounds.theta_min;
		thetaStep = tDiff;
	}else{
		startingTheta = current_bounds.theta_min;
		thetaStep = divs.tDiv;
	}

	//Adjust steps to more tightly fit bounds

	double sum_of_prob = 0;
	int num_part = 0;
	double average_prob = 0;

	for(int i = 0; i < part_array->num_pos; i++){
		if(part_array->positions[i].prob != TRIMMED_PARTICLE_PROB){
			sum_of_prob += part_array->positions[i].prob;
			num_part++;
		}
	}

	average_prob = sum_of_prob/num_part;


	int i = 0;
	int particles_created = 0;

	for(double x = current_bounds.x_min; x <= current_bounds.x_max; x += divs.xDiv){
		for(double y = current_bounds.y_min; y <= current_bounds.y_max; y += divs.yDiv){
			for(double t = startingTheta; t <= current_bounds.theta_max; t += thetaStep){
				while((i < part_array->num_pos)
						&&(part_array->positions[i].prob != TRIMMED_PARTICLE_PROB)){
					i++;
				}

				if(i == part_array->num_pos){
					return;
				}

				part_array->positions[i].x = x;
				part_array->positions[i].y = y;
				part_array->positions[i].theta = t;
				part_array->positions[i].prob = average_prob;
				particles_created++;
			}
		}
	}

	//Deal with leftover particles
	if(particles_created < num_particles){
		thetaStep = tDiff/((num_particles - particles_created) - 1);
		double x = current_bounds.x_min + xDiff/2;
		double y = current_bounds.y_min + yDiff/2;

		for(double t = current_bounds.theta_min; t <= current_bounds.theta_max; t += thetaStep){
			if(particles_created == num_particles){
				return;
			}

			while((i < part_array->num_pos)
					&&(part_array->positions[i].prob != TRIMMED_PARTICLE_PROB)){
				i++;
			}

			if(i == part_array->num_pos){
				return;
			}

			part_array->positions[i].x = x;
			part_array->positions[i].y = y;
			part_array->positions[i].theta = t;
			part_array->positions[i].prob = average_prob;
			particles_created++;
		}
	}
}

divs_struct ParticleFilter::determineDivisions(int num_particles, double xDiff, double yDiff, double tDiff){
	divs_struct divs;
	int numTheta = tDiff/((MAX_THETA - MIN_THETA)/4);

	if(numTheta <= 0){numTheta = 1;}

	divs.tDiv = tDiff/numTheta;

	int remaining_particles = num_particles/numTheta;
	int div_factor = (int)sqrt(remaining_particles);

	divs.xDiv = xDiff/(div_factor - 1);
	divs.yDiv = yDiff/(div_factor - 1);

	return divs;
}

bounds_array ParticleFilter::determineBounds(pos_array high_prob_part){
	bounds_array bounds_set;

	pos_array_set arrays = groupParticles(high_prob_part);

	bounds_set.num_bounds = arrays.num_array;
	bounds_set.bounds_list = (bounds*)malloc(bounds_set.num_bounds*sizeof(bounds));

	for(int i = 0; i < bounds_set.num_bounds; i++){
		bounds_set.bounds_list[i] = determineSetBounds(arrays.arrays[i]);
	}

	//free arrays
	for(int i = 0; i < arrays.num_array; i++){
		free(arrays.arrays[i].positions);
	}

	free(arrays.arrays);

	return bounds_set;
}

bounds ParticleFilter::determineSetBounds(pos_array high_prob_part){
	bounds part_bounds;

	if(high_prob_part.num_pos < 1){
		part_bounds.average_prob = 0;
		part_bounds.theta_max = 0;
		part_bounds.theta_min = 0;
		part_bounds.x_min = 0;
		part_bounds.x_max = 0;
		part_bounds.y_min = 0;
		part_bounds.y_max = 0;

		return part_bounds;
	}

	double prob_sum = 0;

	part_bounds.x_max = part_bounds.x_min = high_prob_part.positions[0].x;
	part_bounds.y_max = part_bounds.y_min = high_prob_part.positions[0].y;
	part_bounds.theta_max = part_bounds.theta_min = high_prob_part.positions[0].theta;

	for(int i = 0; i < high_prob_part.num_pos; i++){
		prob_sum += high_prob_part.positions[i].prob;

		if(part_bounds.x_max < high_prob_part.positions[i].x){
			part_bounds.x_max = high_prob_part.positions[i].x;
		}

		if(part_bounds.x_min > high_prob_part.positions[i].x){
			part_bounds.x_min = high_prob_part.positions[i].x;
		}

		if(part_bounds.y_max < high_prob_part.positions[i].y){
			part_bounds.y_max = high_prob_part.positions[i].y;
		}

		if(part_bounds.y_min > high_prob_part.positions[i].y){
			part_bounds.y_min = high_prob_part.positions[i].y;
		}

		if(part_bounds.theta_max < high_prob_part.positions[i].theta){
			part_bounds.theta_max = high_prob_part.positions[i].theta;
		}

		if(part_bounds.theta_min > high_prob_part.positions[i].theta){
			part_bounds.theta_min = high_prob_part.positions[i].theta;
		}
	}

	part_bounds.average_prob = prob_sum/high_prob_part.num_pos;

	//Add tolerances
	part_bounds.x_max += def.tolerance.point_bound;
	part_bounds.x_min -= def.tolerance.point_bound;
	part_bounds.y_max += def.tolerance.point_bound;
	part_bounds.y_min -= def.tolerance.point_bound;
	part_bounds.theta_max += def.tolerance.theta_bound;
	part_bounds.theta_min -= def.tolerance.theta_bound;

	//Error check x,y
	if(part_bounds.x_max > (FIELD_MM_X/2)){part_bounds.x_max = (FIELD_MM_X/2);}
	if(part_bounds.x_min < (-1*FIELD_MM_X/2)){part_bounds.x_min = (-1*FIELD_MM_X/2);}
	if(part_bounds.y_max > (FIELD_MM_Y/2)){part_bounds.y_max = (FIELD_MM_Y/2);}
	if(part_bounds.y_min < (-1*FIELD_MM_Y/2)){part_bounds.y_min = (-1*FIELD_MM_Y/2);}

	//Error check theta
	if(part_bounds.theta_max > MAX_THETA){part_bounds.theta_max = MAX_THETA;}
	if(part_bounds.theta_min < MIN_THETA){part_bounds.theta_min = MIN_THETA;}

	return part_bounds;
}

/** Splits the give particles into separate groups of particles, and trims the
 * generated group based on probability to a reasonable number
 *
 * @param The particles to split into separate groups
 *
 * @return An array of position arrays which represent the separate groups
 */
pos_array_set ParticleFilter::groupParticles(pos_array part_to_group){
	pos_array* particleGroupOutput;
	pos_array_set temp_return_set;
	temp_return_set.arrays = new pos_array[part_to_group.num_pos];
	temp_return_set.num_array = 0;

	while(part_to_group.num_pos > 0){
		particleGroupOutput = createParticleGroup(part_to_group);

		temp_return_set.num_array++;

		temp_return_set.arrays[temp_return_set.num_array - 1].num_pos = particleGroupOutput[0].num_pos;
		temp_return_set.arrays[temp_return_set.num_array - 1].positions = new pos[temp_return_set.arrays[temp_return_set.num_array - 1].num_pos];

		for(int i = 0; i < particleGroupOutput[0].num_pos; i++){
			temp_return_set.arrays[temp_return_set.num_array - 1].positions[i] = particleGroupOutput[0].positions[i];
		}

		part_to_group.num_pos = particleGroupOutput[1].num_pos;
		part_to_group.positions = new pos[part_to_group.num_pos];

		for(int i = 0; i < particleGroupOutput[1].num_pos; i++){
			part_to_group.positions[i] = particleGroupOutput[1].positions[i];
		}

		free(particleGroupOutput[0].positions);
		free(particleGroupOutput[1].positions);
		free(particleGroupOutput);
	}

	pos_array_set return_set;

	// If lowest is less than half number of particles than highest, eliminate lowest
	if(temp_return_set.num_array > 1){
		pos_array temp;
		double averages[temp_return_set.num_array];
		double temp_avg;

		//Calculate averages
		for(int i = 0; i < temp_return_set.num_array; i++){
			double sum = 0;

			for(int j = 0; j < temp_return_set.arrays[i].num_pos; j++){
				sum += temp_return_set.arrays[i].positions[j].prob;
			}

			averages[i] = sum/temp_return_set.arrays[i].num_pos;
		}

		//Sort the arrays by average probability
		for(int currentPos = 0; currentPos < (temp_return_set.num_array - 1); currentPos++){
			for(int currentCompare = 1; currentCompare < temp_return_set.num_array; currentCompare++){
				if(averages[currentPos] < averages[currentCompare]){
					//Switch positions if less than
					temp_avg = averages[currentPos];
					temp = temp_return_set.arrays[currentPos];

					averages[currentPos] = averages[currentCompare];
					temp_return_set.arrays[currentPos] = temp_return_set.arrays[currentCompare];

					averages[currentCompare] = temp_avg;
					temp_return_set.arrays[currentCompare] = temp;
				}
			}
		}

		//Trim arrays until no more than maximum difference between most and least probable remain
		while(averages[0] > (averages[temp_return_set.num_array-1]+def.tolerance.max_region_prob_dif)){
			temp_return_set.num_array--;
			delete [] temp_return_set.arrays[temp_return_set.num_array].positions;
		}

		return_set.num_array = temp_return_set.num_array;
	}else{
		return_set.num_array = 1;
	}

	//Transfer to malloc'd array
	return_set.arrays = (pos_array*)malloc(temp_return_set.num_array*sizeof(pos_array));

	for(int i = 0; i < temp_return_set.num_array; i++){
		return_set.arrays[i].num_pos = temp_return_set.arrays[i].num_pos;
		return_set.arrays[i].positions = (pos*)malloc(return_set.arrays[i].num_pos*sizeof(pos));

		for(int j = 0; j < return_set.arrays[i].num_pos; j++){
			return_set.arrays[i].positions[j] = temp_return_set.arrays[i].positions[j];
		}

		delete [] temp_return_set.arrays[i].positions;
	}

	delete [] temp_return_set.arrays;

	return return_set;
}

/** Creates a group of particles and splits it into a sub-group and a
 * "remaining" group
 *
 * @param part_to_group The group of particles to split
 *
 * @return Two position arrays. The first (index 0) is a created group of
 * particles. The second if is the particles left to group, or an array with size 0 if none
 * remained
 */
pos_array* ParticleFilter::createParticleGroup(pos_array part_to_group){
	double top_density = -1;
	int top_density_center = 0;
	int top_density_edge = 0;
	int top_density_num_part = 0;
	int current_num_part = 0;
	double current_max_dist = 0;
	double current_calc_dist = 0;
	double current_density = 0;
	double current_max_est_dist = -1;

	//Calculate density of particles at each radius
	if(part_to_group.num_pos > 1){
		for(int current_center = 0; current_center < part_to_group.num_pos; current_center++){
			for(int current_edge = 0; current_edge < part_to_group.num_pos; current_edge++){
				if(current_center != current_edge){
					current_num_part = 2; //center and edge
					//The +1 is a tolerance for double precision (equates to 1mm)
					current_max_dist = calculateDistance(part_to_group.positions[current_center].x,
							part_to_group.positions[current_center].y,
							part_to_group.positions[current_edge].x,
							part_to_group.positions[current_edge].y)
							+ 1;

					//If it is not possible with the maximum dist calculated to achieve a higher density than the
					//current highest, skip this set of checks
					if((current_max_est_dist == -1)||(current_max_est_dist >= current_max_dist)){
						for(int current_check = 0; current_check < part_to_group.num_pos; current_check++){
							if((current_check != current_edge)&&(current_check != current_center)){
								//if distance between this particle and center is <= dist between center and edge, add one
								current_calc_dist = calculateDistance(part_to_group.positions[current_center].x,
										part_to_group.positions[current_center].y,
										part_to_group.positions[current_check].x,
										part_to_group.positions[current_check].y);

								if(current_calc_dist <= current_max_dist){current_num_part++;}
							}
						}

						current_density = current_num_part/(PF_PI*current_max_dist*current_max_dist);

						if(current_density > top_density){
							top_density = current_density;
							top_density_center = current_center;
							top_density_edge = current_edge;
							top_density_num_part = current_num_part;

							//Update the largest allowable radius
							current_max_est_dist = part_to_group.num_pos/top_density;
						}
					}
				}
			}
		}
	}else{
		top_density = 0;
		top_density_center = 0;
		top_density_edge = 0;
		top_density_num_part = 1;
	}

	//create two lists of particles - one of every one within distance, and one of all the others
	pos_array* ret_arrays = (pos_array*)malloc(2*sizeof(pos_array));
	ret_arrays[0].num_pos = top_density_num_part;
	ret_arrays[1].num_pos = part_to_group.num_pos - ret_arrays[0].num_pos;

	ret_arrays[0].positions = (pos*)malloc(ret_arrays[0].num_pos*sizeof(pos));
	ret_arrays[1].positions = (pos*)malloc(ret_arrays[1].num_pos*sizeof(pos));

	int group_index = 0, remaining_index = 0;

	//The +1 is a tolerance for double precision (equates to 1mm)
	current_max_dist = calculateDistance(part_to_group.positions[top_density_center].x,
			part_to_group.positions[top_density_center].y,
			part_to_group.positions[top_density_edge].x,
			part_to_group.positions[top_density_edge].y)
			+ 1;

	for(int i = 0; i < part_to_group.num_pos; i++){
		if((top_density_center == i)||(top_density_edge == i)){
			ret_arrays[0].positions[group_index] = part_to_group.positions[i];
			group_index++;
		}else{
			current_calc_dist = calculateDistance(part_to_group.positions[top_density_center].x,
					part_to_group.positions[top_density_center].y,
					part_to_group.positions[i].x,
					part_to_group.positions[i].y);

			if(current_calc_dist <= current_max_dist){
				ret_arrays[0].positions[group_index] = part_to_group.positions[i];
				group_index++;
			}else{
				ret_arrays[1].positions[remaining_index] = part_to_group.positions[i];
				remaining_index++;
			}
		}
	}

	return ret_arrays;

}

double ParticleFilter::calculateDistance(double x1, double y1, double x2, double y2){
	return sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1));
}

double ParticleFilter::calculateHeading(double x1, double y1, double x2, double y2){
	return atan2((y2-y1), (x2-x1));
}

void ParticleFilter::normalizeParticleProbability(pos_array* part){
	/*double highestAbove = part->positions[0].prob;
	double lowestBelow = part->positions[0].prob;

	for(int i = 1; i < part->num_pos; i++){
		if((part->positions[i].prob != TRIMMED_PARTICLE_PROB)&&(part->positions[i].prob > highestAbove)){
			highestAbove = part->positions[i].prob;
		}else if((part->positions[i].prob != TRIMMED_PARTICLE_PROB)&&(part->positions[i].prob < lowestBelow)){
			lowestBelow = part->positions[i].prob;
		}
	}

	if(lowestBelow < 0.0){
		for(int i = 0; i < part->num_pos; i++){
			part->positions[i].prob -= lowestBelow;
		}

		highestAbove -= lowestBelow;
	}

	if(highestAbove > 1.0){
		for(int i = 0; i < part->num_pos; i++){
			part->positions[i].prob /= highestAbove;
		}
	}*/

	double sum_of_prob = 0;

	for(int i = 0; i < part->num_pos; i++){
		if(part[i].positions[i].prob != TRIMMED_PARTICLE_PROB){
			sum_of_prob += part[i].positions[i].prob;
		}
	}

	for(int i = 0; i < part->num_pos; i++){
		if(part[i].positions[i].prob != TRIMMED_PARTICLE_PROB){
			part[i].positions[i].prob /= sum_of_prob;
		}
	}
}

/*void ParticleFilter::normalizeProbabilityCenter(position_array* part){
	//Compute average probability
	double average = 0;
	int num_not_trim = 0;

	for(int i = 0; i < part->num_positions; i++){
		if(part->positions[i].probability != TRIMMED_PARTICLE_PROB){
			average += part->positions[i].probability;
			num_not_trim++;
		}
	}

	if(num_not_trim > 0){
		average /= num_not_trim;
	}else{
		average = NEW_PARTICLE_PROB;
	}

	if(average < 0){average *= -1;}

	//Compute ratio to make average probability the same as the probability of newly created particles
	double ratio = NEW_PARTICLE_PROB/average;

	//Adjust probabilities
	for(int i = 0; i < part->num_positions; i++){
		if(part->positions[i].probability != TRIMMED_PARTICLE_PROB){
			part->positions[i].probability *= ratio;
		}
	}


	//normalize probabilities
	normalizeParticleProbability(part);
}*/

double ParticleFilter::normalizeTheta(double origTheta){
	while(origTheta < MIN_THETA){
		origTheta += MAX_THETA;
	}

	if(origTheta > MAX_THETA){
		origTheta = fmod(origTheta, MAX_THETA);
	}

	return origTheta;
}

double* ParticleFilter::calculateVariance(pos_array high_prob){
	//Calculate standard div between particle probabilities
	//Average
	double average_prob = 0;
	double average_x = 0;
	double average_y = 0;
	double average_theta = 0;

	double *variances = (double*)malloc(3*sizeof(double));

	for(int i = 0; i < high_prob.num_pos; i++){
		average_prob += high_prob.positions[i].prob;
		average_x += high_prob.positions[i].x;
		average_y += high_prob.positions[i].y;
		average_theta = high_prob.positions[i].theta;
	}

	average_prob /= high_prob.num_pos;
	average_x /= high_prob.num_pos;
	average_y /= high_prob.num_pos;
	average_theta /= high_prob.num_pos;

	//Divs
	double avgdiv_x = 0;
	double avgdiv_y = 0;
	double avgdiv_theta = 0;

	for(int i = 0; i < high_prob.num_pos; i++){
		avgdiv_x += abs(average_x - high_prob.positions[i].x)*high_prob.positions[i].prob;
		avgdiv_y += abs(average_y - high_prob.positions[i].y)*high_prob.positions[i].prob;
		avgdiv_theta += abs(average_theta - high_prob.positions[i].theta)*high_prob.positions[i].prob;
	}

	avgdiv_x /= high_prob.num_pos;
	avgdiv_y /= high_prob.num_pos;
	avgdiv_theta /= high_prob.num_pos;

	variances[0] = (avgdiv_x/(FIELD_MM_X/4))*average_prob;
	variances[1] = (avgdiv_y/(FIELD_MM_Y/4))*average_prob;
	variances[2] = (avgdiv_theta/((MAX_THETA-MIN_THETA)/2))*average_prob;

	return variances;
}

pos_array ParticleFilter::getHighProbParticles(pos_array part, double percentage){
	pos_array high_prob;
	high_prob.num_pos = percentage*part.num_pos;

	if(high_prob.num_pos < 1){
		if(part.num_pos > 1){
			high_prob.num_pos = 1;
		}else{
			return high_prob;
		}
	}

	sortParticles(&part);

	high_prob.positions = (pos*)malloc(high_prob.num_pos*sizeof(Position));

	for(int i = 0; i < high_prob.num_pos; i++){
		high_prob.positions[i] = part.positions[i];
	}

	return high_prob;
}

int ParticleFilter::determineGoalIFF(int postAlignment){
	int goalAlign = P_UNKNOWN;

	double pDist1 = obsPosts[0].distance;
	double pHead1 = obsPosts[0].heading;
	double pDist2;
	double pHead2;
	double goalDist;
	int i;
	int friendly_count = 0;
	int hostile_count = 0;


	if(postAlignment == P_BOTH){
		pDist2 = obsPosts[1].distance;
		pHead2 = obsPosts[1].heading;
		goalDist = (pDist1 + pDist2) / 2.0;

		for (i = 0; i < obsNumRobots; i++){
			if (obsRobotLocations[i].heading < pHead1
					&& obsRobotLocations[i].heading > pHead2
					&& obsRobotLocations[i].distance > (goalDist*.8)
					&& obsRobotLocations[i].distance < (goalDist*1.2)){
				//Friendly goalie seen
				if (((obsRobotTypes[i] == RED_ROBOT && obsTeamRed)
						|| (obsRobotTypes[i] == BLUE_ROBOT && !obsTeamRed))){
					goalAlign = P_FRIENDLY;
					friendly_count++;
				}
				//Enemy goalie seen
				else if (((obsRobotTypes[i] == RED_ROBOT && !obsTeamRed)
						|| (obsRobotTypes[i] == BLUE_ROBOT && obsTeamRed))){
					goalAlign = P_HOSTILE;
					hostile_count++;
				}
			}
		}
	}

	//TODO might be able to filter better using numbers of robots from each team
	if((friendly_count > 0)&&(hostile_count > 0)){
		goalAlign = P_UNKNOWN;
	}

	return goalAlign;
}

lineangle_precalc_data ParticleFilter::lineAngleHeuristicPreCalc(int numEdges, Line lines[MAX_FIELD_EDGES]){
	lineangle_precalc_data data;

	//theta = atan2(-t1/t2)

	data.num_angles = 4*numEdges;
	data.angles = new double[data.num_angles];

	for(int i = 0; i < numEdges; i++){
		double origTheta;

		//Based on trigonometry, atan of the slope is actually the angle the robot is at
		if((lines[i].t1 == 0)||(lines[i].t2 == 0)){
			origTheta = 0;
		}else{
			origTheta = atan2((double)-1*lines[i].t1,lines[i].t2);
		}

		double theta2 = origTheta + PF_PI/2;
		double theta3 = origTheta + PF_PI;
		double theta4 = origTheta + 3*PF_PI/2;

		data.angles[i*4] = normalizeTheta(origTheta);
		data.angles[i*4 + 1] = normalizeTheta(theta2);
		data.angles[i*4 + 2] = normalizeTheta(theta3);
		data.angles[i*4 + 3] = normalizeTheta(theta4);
	}

	data.prob_adjust = def.prob_factor.init;

	//TODO possibly check for "close together" angles (from multiple lines collaborating and
	//giving the same angle as possibility). IF they exist, limit returned angles to those
	//which occur in the list more than once

	return data;
}

void ParticleFilter::lineAngleHeuristic(pos* process, lineangle_precalc_data preCalc){
	for(int i = 0; i < preCalc.num_angles; i++){
		if(preCalc.angles[i] != NAN){
			double headRatio = 0.01;

			//Calculations for heading correctness
			double actualHeading = process->theta;
			double projectedHeading = preCalc.angles[i];

			headRatio = 1 - (2*abs(actualHeading - projectedHeading))/(2*PF_PI);

			double probMod = 1 + ((headRatio - 0.5)*preCalc.prob_adjust);

			if(probMod < 1){probMod *= 1/preCalc.num_angles;}

			process->prob *= probMod;
		}
	}
}

goalheur_precalc_data ParticleFilter::goalPostHeuristicPreCalc(WhichPosts posts, RRCoord* r_coord, bool team_red){
	goalheur_precalc_data precalc_data;

	precalc_data.iff = P_NO_DATA;
	precalc_data.num_data_points = 0;
	precalc_data.prob_adjust = def.prob_factor.init;
	precalc_data.data = NULL;

	//Check for conditions whcih prevent data processing
	if(r_coord == NULL){
		return precalc_data;
	}

	if(posts == pNONE){
		return precalc_data;
	}

	//Determine post position
	int post_pos;

	if((posts == pBLUE_LEFT)||(posts == pYELLOW_LEFT)||(posts == pUNKNOWN_LEFT)){
		post_pos = P_LEFT;
	}else if((posts == pBLUE_RIGHT)||(posts == pYELLOW_RIGHT)||(posts == pUNKNOWN_RIGHT)){
		post_pos = P_RIGHT;
	}else if((posts == pBLUE_BOTH)||(posts == pYELLOW_BOTH)||(posts == pUNKNOWN_BOTH)){
		post_pos = P_BOTH;
	}else{
		post_pos = P_EITHER;
	}

	precalc_data.iff = determineGoalIFF(post_pos);

	//Process left then right for the "both" case, as this is the order of the RRCoord array
	if((post_pos == P_LEFT)||(post_pos == P_RIGHT)){
		double hostile_x = 0;
		double hostile_y = 0;
		double friendly_x = 0;
		double friendly_y = 0;

		if(post_pos == P_LEFT){
			hostile_x = -700;
			hostile_y = 3000;
			friendly_x = 700;
			friendly_y = -3000;
		}else{
			hostile_x = 700;
			hostile_y = 3000;
			friendly_x = -700;
			friendly_y = -3000;
		}

		if(precalc_data.iff == P_UNKNOWN){
			precalc_data.data = (pos_location*)malloc(2*sizeof(pos_location));
			precalc_data.num_data_points = 2;
		}else{
			precalc_data.data = (pos_location*)malloc(sizeof(pos_location));
			precalc_data.num_data_points = 1;
		}

		int index = 0;

		//One post, lessen impact on probability
		precalc_data.prob_adjust *= def.prob_factor.one_post;

		if((precalc_data.iff == P_HOSTILE)||(precalc_data.iff == P_UNKNOWN)){
			precalc_data.data[index].x = hostile_x;
			precalc_data.data[index].y = hostile_y;
			precalc_data.data[index].distance = r_coord[0].distance;
			precalc_data.data[index].heading = r_coord[0].heading;
			index++;
		}

		if((precalc_data.iff == P_FRIENDLY)||(precalc_data.iff == P_UNKNOWN)){
			precalc_data.data[index].x = friendly_x;
			precalc_data.data[index].y = friendly_y;
			precalc_data.data[index].distance = r_coord[0].distance;
			precalc_data.data[index].heading = r_coord[0].heading;
			index++;
		}
	}else{
		//Both and either are similar enough that they can be handled by nearly identical code
		//Mod to probability adjust for unknown is already done when iff is set
		if(precalc_data.iff == P_UNKNOWN){
			precalc_data.data = (pos_location*)malloc(4*sizeof(pos_location));
			precalc_data.num_data_points = 4;
		}else{
			precalc_data.data = (pos_location*)malloc(2*sizeof(pos_location));
			precalc_data.num_data_points = 2;
		}

		int index = 0;

		if((precalc_data.iff == P_HOSTILE)||(precalc_data.iff == P_UNKNOWN)){
			precalc_data.data[index].x = -700;
			precalc_data.data[index].y = 3000;
			index++;
			precalc_data.data[index].x = 700;
			precalc_data.data[index].y = 3000;
			index++;
		}

		if((precalc_data.iff == P_FRIENDLY)||(precalc_data.iff == P_UNKNOWN)){
			precalc_data.data[index].x = 700;
			precalc_data.data[index].y = -3000;
			index++;
			precalc_data.data[index].x = -700;
			precalc_data.data[index].y = -3000;
		}

		//Do double heading dist, will be replaced if both are seen
		precalc_data.data[0].distance = r_coord[0].distance;
		precalc_data.data[0].heading = r_coord[0].heading;

		precalc_data.data[1].distance = r_coord[0].distance;
		precalc_data.data[1].heading = r_coord[0].heading;

		if(precalc_data.iff == P_UNKNOWN){
			precalc_data.data[2].distance = r_coord[0].distance;
			precalc_data.data[2].heading = r_coord[0].heading;

			precalc_data.data[3].distance = r_coord[0].distance;
			precalc_data.data[3].heading = r_coord[0].heading;
		}

		if(post_pos == P_BOTH){
			//Check for valid triangle and lessen impact on probability if not present
			if(abs(r_coord[0].distance - r_coord[1].distance) > 1400){
				//invalid triangle
				precalc_data.prob_adjust *= def.prob_factor.invalid_triangle;
			}

			precalc_data.data[1].distance = r_coord[1].distance;
			precalc_data.data[1].heading = r_coord[1].heading;

			if(precalc_data.iff == P_UNKNOWN){
				precalc_data.data[3].distance = r_coord[1].distance;
				precalc_data.data[3].heading = r_coord[1].heading;
			}
		}else{
			//Either post, lessen impact on probability
			precalc_data.prob_adjust *= def.prob_factor.either_post;
		}
	}

	return precalc_data;
}

void ParticleFilter::goalPostHeuristic(pos* process, goalheur_precalc_data preCalc){
	double distance, heading;

	//Exit this heuristic if no goals found
	if(preCalc.iff == 0){
		return;
	}

	for(int i = 0; i < preCalc.num_data_points; i++){
		if(preCalc.data != NULL){
			//Normalize theta
			process->theta = normalizeTheta(process->theta);

			//Heavily modify probability based on general direction faced
			if(preCalc.iff == P_FRIENDLY){
				if(process->theta > ((MAX_THETA-MIN_THETA)/2)){
					process->prob *= (2 - def.prob_factor.direction);
				}else{
					process->prob *= def.prob_factor.direction;
				}
			}else if(preCalc.iff == P_HOSTILE){
				if(process->theta > ((MAX_THETA-MIN_THETA)/2)){
					process->prob *= def.prob_factor.direction;
				}else{
					process->prob *= (2 - def.prob_factor.direction);
				}

			}

			//Calculate distance and heading from given point
			distance = calculateDistance(preCalc.data[i].x,preCalc.data[i].y,process->x,process->y);
			heading = calculateHeading(process->x,process->y, preCalc.data[i].x,preCalc.data[i].y);

			//Make sure to mod 2 PI the heading difference
			heading = normalizeTheta(heading);

			double distRatio = 1 - (2*abs(preCalc.data[i].distance - distance))/sqrt(1000*(2*FIELD_MM_X/1000*FIELD_MM_X + 2*FIELD_MM_Y/1000*FIELD_MM_Y));
			double headRatio = 0.01;

			//Calculations for heading correctness
			if(heading != NAN){
				//Converts both the projected heading of the particle and the calculated heading
				//of the two points (particle and absolute of pole)
				double actualHeading = normalizeTheta(heading);
				double projectedHeading = normalizeTheta(preCalc.data[i].heading + process->theta);

				headRatio = 1 - (2*abs(actualHeading - projectedHeading))/(2*PF_PI);
			}

			double prob_mod = 1 + ((distRatio*headRatio - 0.5)*preCalc.prob_adjust);

			process->prob *= prob_mod;
		}
	}
}

void ParticleFilter::odometryAdjustment(Odometry newOdo, pos_array* part){
	Odometry changeInOdo;

	//Correct Odometry
	newOdo.forward = correctForward(newOdo.forward);
	newOdo.left = correctLeft(newOdo.left);
	newOdo.turn = correctTurn(newOdo.turn);

	//Calculate the delta of the position according to odometry
	changeInOdo.forward = newOdo.forward - prevOdo.forward;
	changeInOdo.left = newOdo.left - prevOdo.left;
	changeInOdo.turn = newOdo.turn - prevOdo.turn;

	//Update the "previous" odometry to the current data
	prevOdo.forward = newOdo.forward;
	prevOdo.left = newOdo.left;
	prevOdo.turn = newOdo.turn;

	//Preloading to adjust odometry for heading of robot
	double xChange, yChange, radius, angleWithoutHeading;

	radius = sqrt(changeInOdo.forward*changeInOdo.forward + changeInOdo.left*changeInOdo.left);
	angleWithoutHeading = atan2(changeInOdo.forward, -1*changeInOdo.left);
	angleWithoutHeading = normalizeTheta(angleWithoutHeading);

	//Apply change in x,y, and theta to particles
	//NOTE: -1*changeInOd.left is to account for the reversed frame of reference
	//Odometry uses
	for(int i = 0; i < part->num_pos; i++){
		xChange = radius*cos(angleWithoutHeading + part->positions[i].theta);
		yChange = radius*sin(angleWithoutHeading + part->positions[i].theta);
		part->positions[i].x += xChange;
		part->positions[i].y += yChange;
		part->positions[i].theta += changeInOdo.turn;

		//Adjust for slight uncertainty added by using odometry
		//part->positions[i].probability *= def.prob_factor.odometry;
	}
}
