/*
 * ParticleFilter.hpp
 *
 *  Created on: Feb 5, 2012
 *      Author: Ryan
 */

#ifndef PARTICLEFILTER_HPP_
#define PARTICLEFILTER_HPP_

#include "pos.hpp"
#include "utils/Odometry.hpp"
#include "perception/vision/VisionDefs.hpp"
#include "utils/RRCoord.hpp"
#include "perception/localisation/Localiser.hpp"
#include "pf_def.hpp"
#include "bounds.hpp"

//Involatile values - these are not "defaults", but facts of the environment/system,
//and may be used directly
/** The default number of particles maintained by the particle filter */
const int NUM_PARTICLES = 120;

/** The probability assigned to newly created particles */
//const double NEW_PARTICLE_PROB = 0.5;
/** Probability assigned to particles that are going to be replaced */
const double TRIMMED_PARTICLE_PROB = -1.0;

/** Internal constant - value of PI. math.h provided constant was not working,
 * and uses more precision than necessary for this application */
const double PF_PI = 3.14159;

/** The minimum value for a theta entry */
const double MIN_THETA = 0;
/** The maximum value for a theta entry */
const double MAX_THETA = 2*PF_PI;

/** The x-dimension of the field, in mm */
const double FIELD_MM_X = 5400;
/** The y-dimension of the field, in mm */
const double FIELD_MM_Y = 7400;

/** Constant that indicates not data was available this loop */
const int P_NO_DATA = 0;
/** Constant that indicates that the friendly end of the field is being faced */
const int P_FRIENDLY = 1;
/** Constant that indicates that the hostile end of the field is being faced */
const int P_HOSTILE	= 2;
/** Constant that indicates that the end of the field being faced is unknown */
const int P_UNKNOWN = 3;
/** Constant that indicates that the left post is being viewed */
const int P_LEFT = 3;
/** Constant that indicates that the right post is being viewed */
const int P_RIGHT = 4;
/** Constant that indicates that both posts are being viewed */
const int P_BOTH = 5;
/** Constant that indicates that the post being viewed could be left or right */
const int P_EITHER = 6;

//"Volatile" constants - these are defaults for the system, and may be different at runtime based on
//input. Should be referred to through the particle filter definition structure
/** The percentage of particles to replace each loop */
const double DEFAULT_TURNOVER_PERCENTAGE = 0.4;
/** The percentage of particles to treat as high probability */
const double DEFAULT_HIGH_PROB_PERCENTAGE = 0.45;

/** The tolerance to expand boundaries by for x and y */
const double DEFAULT_POINT_TOL = 50.0;
/** The tolerance to expand boundaries by for theta */
const double DEFAULT_THETA_TOL = 0.1;
/** The maximum difference between avergae probabilities for two regions */
const double DEFAULT_MAX_REGION_PROB_DIF = 0.25;

/** The initial probability adjustment factor */
const double DEFAULT_INIT_PROB_ADJUST = 1;//0.0625;
/** Adjustment to probability slow down if only one post is seen */
const double DEFAULT_ONE_POST_PROB_ADJUST = 0.75;
/** Adjustment to probability slow down if one post is seen, unknown side */
const double DEFAULT_EITHER_POST_PROB_ADJUST = 0.5;
/** Adjustment to probability slow down if both post are seen, but an invalid
 * triangle is formed by the distances*/
const double DEFAULT_INVALID_TRI_PROB_ADJUST = 0.0625;
/** Adjustment to the probability to slow down if it is not known which posts are seen */
const double DEFAULT_UNKNOWN_POSTS_PROB_ADJUST = 0.25;
/** Constant that defines the probability modification for facing the
 * correct/incorrect general direction of seen posts */
const double DEFAULT_DIRECTION_PROB_MOD = 0.0625;
/** Constant that lowers probability based on odometry use */
const double DEFAULT_ODOMETRY_PROB_MOD = 0.98;

typedef struct{
	double xDiv, yDiv, tDiv;
}divs_struct;

typedef struct{
	double distance, heading;
	double x,y;
}pos_location;

typedef struct{
	int iff;
	int num_data_points;
	double prob_adjust;
	pos_location* data;
}goalheur_precalc_data;

typedef struct{
	int num_angles;
	double* angles;
	double prob_adjust;
}lineangle_precalc_data;

typedef struct{
	pos_array* arrays;
	int num_array;
}pos_array_set;


class ParticleFilter : public Localiser{
	friend class Localiser;

public:
	/** Creates a ParticleFilter with the default number of particles. The
	 * filter will also create its own instance of a DebuggingTools object*/
	ParticleFilter();

	/** Creates a ParticleFilter with the designated number of particles. The
	 * filter will create its own instance of a DebuggingTools object
	 *
	 * @param num_particles The number of particles that the filter should
	 * maintain during operation */
	ParticleFilter(int num_particles);

	~ParticleFilter();

	/** Runs algorithms to generate possible points the robot be be located at */
	Position localise();

	/** Updates positions with new sensor data */
	Position update();
protected:
	/** Array of positions maintained by the filter for processing */
	pos_array particles;
	/** Record of previous odometry to allow relative processing */
	Odometry prevOdo;
	/** Defines the parameters of the particle filter */
	pf_def def;
	/** Keeps a record of the original values in case they need to be reset after adjustment */
	pf_def orig_def;

	bool penalizeProc;

	/**	Adjusts the given positions for changes in position given by sources
	 * such as odometry. Adjusts x, y, and theta
	 *
	 * @param part The array of Positions to adjust */
	void adjustParticles(pos_array* part);

	/** Evaluates sensor data and interprets it to adjust particle probabilities.
	 *
	 * @param part The particles to change
	 *
	 * @return true if changes were made, false otherwise */
	bool evaluateParticles(pos_array* part);

	/** Trims a given percentage of particles from the provided array of positions.
	 * Trimming consists of changing the probability on the particles to a set value
	 * which marks them as "to be replaced"
	 *
	 * @param part The particle array to trim
	 * @param percentage The percentage of particles to trim */
	void trimParticles(pos_array* part, double percentage);

	/** Replaces all trimmed particles in the given array with new particles, in the area of
	 * the highest probability particles
	 *
	 * @param part The array of positions to process */
	void multiplyParticles(pos_array* part);


	/** Sorts the given particle array by probability, greatest at index 0. If the array is already
	 * sorted, then the function exits
	 *
	 * @param part A pointer to the array to sort*/
	void sortParticles(pos_array* part);

	/** Determines if the given position_array is sorted by probability (greatest at index 0,
	 * least at last index)
	 *
	 * @param part The array to analyze
	 *
	 * @return true if sorted, false otherwise */
	bool sorted(pos_array part);

	/** Initializes the filter. Sets up particle array, populates with intial particles
	 *
	 * @param num_particles The number of particles to maintain, or -1 if default
	 * @param debug_tools The reference to a DebuggingTools object to use, or NULL
	 * if a new reference should be created */
	void initialize(int num_particles, pf_def* i_def);

	/** Determines if a position is within the x,y bounds of the field
	 *
	 * @param pos The Position to test
	 *
	 * @return true if in bounds, false otherwise
	 */
	bool inBounds(pos position);

	/** Replaces all particles marked to be trimmed with new particles near the highest
	 * probability particles
	 *
	 * @param num_particles The number of particles marked as trimmed
	 * @param bounds_to_fill A set of bounds to generate the new particles within */
	void createParticles(int num_particles, pos_array* part_array, bounds_array bounds_to_fill);

	/** Determines which divisions between x, y, and theta points to use when generating new
	 * points
	 *
	 * @param num_particles The number of particles which are being regenerated
	 * @param xDiff The difference between minimum and maximum x
	 * @param yDiff The difference between minimum and maximum y
	 * @param tDiff The difference between minimum and maximum theta
	 *
	 * @return An array of the divisions for x, y, and theta*/
	divs_struct determineDivisions(int num_particles, double xDiff, double yDiff, double tDiff);

	/** Determines the boundaries to create new particles within, given a set
	 * of the highest probability particles
	 *
	 * @param high_prob_part An array of the highest probability particles
	 *
	 * @return An array of bounds to fill */
	bounds_array determineBounds(pos_array high_prob_part);

	/** Determines the bounds for a single set of particles
	 *
	 * @param high_prob_part An array of high probability particles
	 * to generate boundaries for
	 *
	 * @return The boundaries for this set of particles */
	bounds determineSetBounds(pos_array high_prob_part);

	/** Splits the give particles into separate groups of particles, and trims the
	 * generated group based on probability to a reasonable number
	 *
	 * @param The particles to split into separate groups
	 *
	 * @return An array of position arrays which represent the separate groups
	 */
	pos_array_set groupParticles(pos_array part_to_group);

	/** Creates a group of particles and splits it into a sub-group and a
	 * "remaining" group
	 *
	 * @param part_to_group The group of particles to split
	 *
	 * @return Two position arrays. The first (index 0) is a created group of
	 * particles. The second if is the particles left to group, or NULL if none
	 * remained
	 */
	pos_array* createParticleGroup(pos_array part_to_group);

	/** Determines the distance between two points
	 *
	 * @param x1, y1 The first point
	 * @param x2, y2 The second point
	 *
	 * @return The distance between the two points */
	double calculateDistance(double x1, double y1, double x2, double y2);

	/** Determines the heading from one point to another
	 *
	 * @param x1, y1 The first point
	 * @param x2, y2 The second point
	 *
	 * @return The heading from point one to point two */
	double calculateHeading(double x1, double y1, double x2, double y2);

	/** Normalizes the particle probabilities to between 1 and 0 */
	void normalizeParticleProbability(pos_array* part);

	/** Centers the probability of all particles around the probability given to
	 * new particles */
	//void normalizeProbabilityCenter(pos_array* part);

	/** Puts theta in range of (MIN_THETA, MAX_THETA) */
	double normalizeTheta(double origTheta);

	/** Calculates the variance between high prob particles to determine accuracy of filter */
	double* calculateVariance(pos_array high_prob);

	/** Finds the highest probability particles in the given array (total
	 * is the given percentage of the provided array)
	 *
	 * @param part THe particle array to select particles from
	 * @param percentage A value in the bounds(0.0, 1.0] which is the percentage
	 * of particles to select and include in the generated array
	 *
	 * @return A position_array of selected particles
	 */
	pos_array getHighProbParticles(pos_array part, double percentage);

	//TODO
	int determineGoalIFF(int postAlignment);

	/** Processes sensor data pertaining to the field edges and their slope. Converts this information
	 * to possible angles for the robot's position
	 *
	 * @param numEdges The number of edges seen
	 * @param lines An array of the field edges represented as Line objects
	 * @return A structure containing angle and probability information
	 */
	lineangle_precalc_data lineAngleHeuristicPreCalc(int numEdges, Line lines[MAX_FIELD_EDGES]);

	/** Processes generated data, using possible robot angular positions as a filter/probability
	 * adjustment
	 *
	 * @param process The position to process with the edge information
	 * @param preCalc The information generated by pre-heuristic algorithms pertaining to
	 * field edges
	 */
	void lineAngleHeuristic(pos* process, lineangle_precalc_data preCalc);

	/** Processes sensor input and converts it into a format easier to process by the particle filter
	 *
	 * @param posts A constant representing which posts were seen, and what team they belong to
	 * @param r_coord An array of robot-relative coordinates to the seen posts
	 * @param team_red A boolean indicating which team the robot is on. true if on "red" team, false if
	 * on "blue" team
	 *
	 * @return A structure containing processed data about post locations and which side the posts are on */
	goalheur_precalc_data goalPostHeuristicPreCalc(WhichPosts posts, RRCoord* r_coord, bool team_red);

	/** Takes in pre-calculated data from sensors about goal posts.
	 *
	 * @param process A pointer to the Position to modify
	 * @param preCalc Data to use when modifying probability */
	void goalPostHeuristic(pos* process, goalheur_precalc_data preCalc);

	/** Adjusts the given particles using the provided odometry data */
	void odometryAdjustment(Odometry newOdo, pos_array* part);
};

#endif /* PARTICLEFILTER_HPP_ */
